from pathlib import Path
from typing import Optional
from pprint import pprint
from flask import render_template_string, send_file

from src.rendering import GENERIC_FILE_MAPPING
from src.rendering.markdown import render_markdown


def count_file_extensions(path):
    files = path.glob("*")
    files_map = {}
    for file in files:
        if file.is_file():
            extension = file.suffix.lower()
            if extension in files_map:
                files_map[extension] += 1
            else:
                files_map[extension] = 1
    return files_map


def determine_type(path: Path) -> tuple[str, str, Optional[str]]:
    if path.is_file():
        # Find extension in GENERIC_FILE_MAPPING, it may match multiple
        generic_mapping = []
        for file_type, extensions in GENERIC_FILE_MAPPING.items():
            if path.suffix[1:] in extensions:
                if not generic_mapping:
                    generic_mapping = [file_type]
                else:
                    generic_mapping.append(file_type)
        generic_mapping.reverse()
        if not generic_mapping:
            generic_mapping = ["other"]
        extension = path.suffix[1:]
        return "file", generic_mapping, extension
    # Directory
    files_map = count_file_extensions(path)
    if files_map:
        most_seen_extension = max(files_map, key=files_map.get)
        generic_mapping = []
        for file_type, extensions in GENERIC_FILE_MAPPING.items():
            if most_seen_extension[1:] in extensions:
                if generic_mapping is None:
                    generic_mapping = [file_type]
                else:
                    generic_mapping.append(file_type)
        generic_mapping.reverse()
        if generic_mapping is None:
            generic_mapping = ["other"]
        return "folder", generic_mapping, most_seen_extension[1:]
    else:
        return "folder", [], None


DEFAULT_ERROR_TEMPLATE = """
<div class="content">
    <h1>Error: {{ code }}</h1>
    <p>{{ message }}</p>
    <p>{{ description }}</p>
    <p>PS: If you are a theme developer, or the owner of this site, you can style this page with a __error.html file</p>
</div>
"""


def render_error_page(
    error_code: int,
    error_message: str,
    error_description: str,
    template_path: Path = Path("./"),
):
    inp = DEFAULT_ERROR_TEMPLATE
    if (template_path / "__error.html").exists():
        inp = (template_path / "__error.html").read_text()
    content = render_template_string(
        inp,
        error_code=error_code,
        error_message=error_message,
        error_description=error_description,
    )
    return (
        render_template_string(
            (template_path / "base.html").read_text(),
            content=content,
            styles=["/base.css", "/__error.css"],
        ),
        error_code,
    )


def render_page(
    path: Path,
    base_path: Path = Path("./"),
    template_path: Path = Path("./"),
    style_path: Path = Path("./"),
):
    """
    Renders a web page based on the provided path and optional base, template, and style paths.

    Args:
        path (Path): The path to the target file or directory to render.
        base_path (Path, optional): The base path to use for relative paths. Defaults to Path("./").
        template_path (Path, optional): The path to the directory containing HTML templates. Defaults to Path("./").
        style_path (Path, optional): The path to the directory containing CSS styles. Defaults to Path("./").

    Returns:
        str: The rendered HTML content of the page, or an error page if the target path does not exist or no suitable template is found.

    Raises:
        Exception: If the base template (base.html) is not found in the template_path.

    Notes:
        - If the target path does not exist, a 404 error page is rendered.
        - If the target path is a file, the function attempts to determine its type, category, and extension.
        - The function generates a list of possible CSS styles based on the target path and its type/category.
        - The function searches for suitable HTML templates based on the target path and its type/category.
        - If no suitable template is found, the function either sends the file directly (if it's a file) or renders a 404 error page.
        - If the target file is a document, its content is rendered as Markdown.
        - The function ensures that the base template (base.html) exists before rendering the final content.
    """
    if not path.exists():
        return render_error_page(
            error_code=404,
            error_message="Not Found",
            error_description="The requested resource was not found on this server.",
            template_path=template_path,
        )
    target_path = path
    target_file = path
    if path.is_file():
        target_path = path.parent
    type, category, extension = determine_type(target_file)

    relative_path = target_file.relative_to(base_path)
    relative_dir = target_path.relative_to(base_path)

    """
    The styles are ordered in the following manner:

    Specific style for the target path (e.g., /path/to/target.css).
    Specific styles for the type and extension in the current and parent directories
        (e.g., /path/to/__file.html.css).
    Specific styles for the type and category in the current and parent directories
        (e.g., /path/to/__file.document.css).
    Base style (/base.css).
    This ordering ensures that the most specific styles are applied first, followed by
        more general styles, and finally the base style.
    """
    styles = []
    styles.append("/" + str(relative_path) + ".css")

    search_path = style_path / relative_dir
    while search_path >= style_path:
        if (search_path / f"__{type}.{extension}.css").exists():
            styles.append(
                "/"
                + str(search_path.relative_to(style_path))
                + f"/__{type}.{extension}.css"
            )
        for c in reversed(category):
            if (search_path / f"__{type}.{c}.css").exists():
                styles.append(
                    "/"
                    + str(search_path.relative_to(style_path))
                    + f"/__{type}.{c}.css"
                )
        search_path = search_path.parent

    styles.append("/base.css")

    styles = [t for t in styles if (style_path / t[1:]).exists()]

    templates = []
    if type == "folder":
        if (template_path / relative_dir / "__folder.html").exists():
            templates.append(relative_dir / "__folder.html")
    else:
        if (template_path / (str(relative_path) + ".html")).exists():
            templates.append(template_path / (str(relative_path) + ".html"))

    if len(templates) == 0:
        search_path = template_path / relative_dir
        while search_path >= template_path:
            if (search_path / f"__{type}.{extension}.html").exists():
                templates.append(search_path / f"__{type}.{extension}.html")
                break
            for c in reversed(category):
                if (search_path / f"__{type}.{c}.html").exists():
                    templates.append(search_path / f"__{type}.{c}.html")
                    break
            search_path = search_path.parent

    if len(templates) == 0:
        if type == "file":
            return send_file(target_file)
        else:
            return render_error_page(
                404,
                "Not Found",
                "The requested resource was not found on this server.",
                template_path,
            )

    content = ""
    c_frontmatter = None
    if "document" in category and type == "file":
        content, c_frontmatter, obj = render_markdown(target_file)

    if not (template_path / "base.html").exists():
        raise Exception("Base template not found")


    # The first found template is the most specific one for the content.
    page_template_path = templates[0]

    template_vars = {
        "content": content,
        "styles": styles,
        "currentPath": str(relative_path),
        "metadata": c_frontmatter if "document" in category and type == "file" else None,
    }

    # First, render the specific page template.
    final_content = render_template_string(
        page_template_path.read_text(), **template_vars
    )

    # Now, render the base template, providing the result of the page
    # template as the 'content' variable.
    template_vars["content"] = final_content
    return render_template_string(
        (template_path / "base.html").read_text(), **template_vars
    )
